/*********************************************************************************
*                     Bomben-Spiel
*                   ================
*                   Uwe Berger; 2009
*
*
* ---------
* Have fun!
*
**********************************************************************************/
#ifndef F_CPU
	#define F_CPU 1000000
#endif

#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <stdio.h>


#include "font.h"

//*** Hardwaredefinitionen ***
// Matrix-Zeilen
#define ROW_0_DDR		DDRC
#define ROW_0_PORT		PORTC
#define ROW_0			PC0
#define ROW_1_DDR		DDRC
#define ROW_1_PORT		PORTC
#define ROW_1			PC1
#define ROW_2_DDR		DDRC
#define ROW_2_PORT		PORTC
#define ROW_2			PC2
#define ROW_3_DDR		DDRC
#define ROW_3_PORT		PORTC
#define ROW_3			PC3
#define ROW_4_DDR		DDRC
#define ROW_4_PORT		PORTC
#define ROW_4			PC4
#define ROW_5_DDR		DDRC
#define ROW_5_PORT		PORTC
#define ROW_5			PC5
#define ROW_6_DDR		DDRD
#define ROW_6_PORT		PORTD
#define ROW_6			PD0
#define ROW_7_DDR		DDRD
#define ROW_7_PORT		PORTD
#define ROW_7			PD1
// Matrix-Spalten
#define COL_0_DDR		DDRD
#define COL_0_PORT		PORTD
#define COL_0			PD3
#define COL_1_DDR		DDRD
#define COL_1_PORT		PORTD
#define COL_1			PD4
#define COL_2_DDR		DDRD
#define COL_2_PORT		PORTD
#define COL_2			PD5
#define COL_3_DDR		DDRD
#define COL_3_PORT		PORTD
#define COL_3			PD6
#define COL_4_DDR		DDRD
#define COL_4_PORT		PORTD
#define COL_4			PD7
#define COL_5_DDR		DDRB
#define COL_5_PORT		PORTB
#define COL_5			PB0
#define COL_6_DDR		DDRB
#define COL_6_PORT		PORTB
#define COL_6			PB1
#define COL_7_DDR		DDRB
#define COL_7_PORT		PORTB
#define COL_7			PB2

// Taster
#define TASTER_ON_OFF_DDR	DDRD
#define TASTER_ON_OFF_PORT	PORTD
#define TASTER_ON_OFF_PINX  PIND
#define TASTER_ON_OFF_PIN	PD2

#define TASTER_LEFT_DDR		DDRB
#define TASTER_LEFT_PORT	PORTB
#define TASTER_LEFT_PINX	PINB
#define TASTER_LEFT_PIN		PB7

#define TASTER_RIGHT_DDR	DDRB
#define TASTER_RIGHT_PORT	PORTB
#define TASTER_RIGHT_PINX	PINB
#define TASTER_RIGHT_PIN	PB6

#define DEBOUNCE_MS			75

// Tastencodes
#define ON_OFF	1
#define LEFT	2
#define RIGHT	3

// ein paar Dimensionen
#define MATRIX_DX		8	// Breite LED-Matrix
#define MATRIX_DY		8	// Hoehe LED-Matrix
#define CURSOR_SIZE		2	// Groesse Cursor/Schlaeger
#define MAX_BOMBS		3	// Anzahl gleichzeitiger Bomben
#define SPEED_NUMBERS	8	// Anzahl unterschiedlicher Geschwindigkeiten


volatile uint8_t row = 0;
volatile int8_t cx = (MATRIX_DX/2)-(CURSOR_SIZE/2);	// Cursor in Mittelstellung
uint8_t next_step = 0;
uint8_t step_counter = 0;

#define PRELOAD_T0 			217
#define STEP_COUNTER_MAXVAL	75


volatile uint8_t led[MATRIX_DY] = {
							0b10101010,
							0b01010101,
							0b10101010,
							0b01010101,
							0b10101010,
							0b01010101,
							0b10101010,
							0b01010101
							};
							
volatile uint8_t bomb[MAX_BOMBS][4];	// 0-> x
										// 1-> moment. y 
										// 2-> moment. Zeit-Incr.
										// 3-> initiales Zeit-Incr.
										
volatile uint16_t lost_bomb = 0;
volatile uint16_t win_bomb = 0;
										

//************************************************************************
void matrix_clear(void) {
	// alle Zeilen aus (Low)
	ROW_0_PORT &= ~(1<<ROW_0);
	ROW_1_PORT &= ~(1<<ROW_1);
	ROW_2_PORT &= ~(1<<ROW_2);
	ROW_3_PORT &= ~(1<<ROW_3);
	ROW_4_PORT &= ~(1<<ROW_4);
	ROW_5_PORT &= ~(1<<ROW_5);
	ROW_6_PORT &= ~(1<<ROW_6);
	ROW_7_PORT &= ~(1<<ROW_7);
	// alle Spalten aus (High)
	COL_0_PORT |= (1<<COL_0);
	COL_1_PORT |= (1<<COL_1);
	COL_2_PORT |= (1<<COL_2);
	COL_3_PORT |= (1<<COL_3);
	COL_4_PORT |= (1<<COL_4);
	COL_5_PORT |= (1<<COL_5);
	COL_6_PORT |= (1<<COL_6);
	COL_7_PORT |= (1<<COL_7);
}

//************************************************************************
void set_led(uint8_t x, uint8_t y, uint8_t val)
{
	x = 7-x;
	if (val) {led[y] |= (1 << x);} else {led[y] &= ~(1 << x);}
}


//************************************************************************
uint8_t get_led(uint8_t x, uint8_t y)
{
	return (bit_is_set(led[y], MATRIX_DX-1-x));
}


//************************************************************************
void draw_cursor(int8_t delta) {
	uint8_t i;
	// alten Cursor loeschen
	for (i=cx; i<(cx+CURSOR_SIZE); i++) set_led(i, MATRIX_DY-1, 0);
	// neue Cursorposition berechnen
	cx = cx + delta;
	if (cx>(MATRIX_DX-CURSOR_SIZE)) cx=MATRIX_DX-CURSOR_SIZE;
	if (cx<0) cx=0;
	// neuen Cursor zeichen
	for (i=cx; i<(cx+CURSOR_SIZE); i++) set_led(i, MATRIX_DY-1, 1);
}

//************************************************************************
void clear_all(void)
{
	uint8_t i;
	for (i=0; i<MATRIX_DX; i++) led[i] = 0;
}
//************************************************************************
ISR(TIMER0_OVF_vect)
{
	TCNT0 = PRELOAD_T0;	
	// Matrix zuruecksetzen
	matrix_clear();
	// naechste Zeile/Spalte ermitteln
	row++;
	if (row > 7) row = 0;
	// Spalten setzen
	if (get_led(0, row)) COL_0_PORT &= ~(1<<COL_0);
	if (get_led(1, row)) COL_1_PORT &= ~(1<<COL_1);
	if (get_led(2, row)) COL_2_PORT &= ~(1<<COL_2);
	if (get_led(3, row)) COL_3_PORT &= ~(1<<COL_3);
	if (get_led(4, row)) COL_4_PORT &= ~(1<<COL_4);
	if (get_led(5, row)) COL_5_PORT &= ~(1<<COL_5);
	if (get_led(6, row)) COL_6_PORT &= ~(1<<COL_6);
	if (get_led(7, row)) COL_7_PORT &= ~(1<<COL_7);
	// naechste Zeile ein (Low)
	switch (row) {
		case 0: ROW_0_PORT |= (1 << ROW_0); break;
		case 1: ROW_1_PORT |= (1 << ROW_1); break;
		case 2: ROW_2_PORT |= (1 << ROW_2); break;
		case 3: ROW_3_PORT |= (1 << ROW_3); break;
		case 4: ROW_4_PORT |= (1 << ROW_4); break;
		case 5: ROW_5_PORT |= (1 << ROW_5); break;
		case 6: ROW_6_PORT |= (1 << ROW_6); break;
		case 7: ROW_7_PORT |= (1 << ROW_7); break;
	}
	// step_counter/nextstep hochzaehlen/setzen
	step_counter++;
	if (step_counter > STEP_COUNTER_MAXVAL) {
		step_counter = 0;
		next_step = 1;
	}
}

//************************************************************************
ISR(INT0_vect)
{
	// deklariert, um wieder aus Sleep-Mode zu kommen
}

//************************************************************************
void my_delay_ms(uint16_t ms)
{
	uint16_t i;
	for (i=0; i<ms; i++) _delay_ms(1);
}

//************************************************************************
void go_sleep(void)
{
	// Timer0-OVF ausschalten
	TIMSK  &= ~(1 << TOIE0);
	// Ports entsprechend "loeschen"...
	matrix_clear();
	// Pause damit nicht gleich INT0 
	my_delay_ms(500);
	// INT0 einschalten
	GICR  |= (1<<INT0);
	// Sleep-Mode einstellen
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);	
	// Sleep-Mode einschalten
	sleep_mode();
	// INT0 ausschalten		
	GICR  &= ~(1<<INT0);
	//	Timer0-OVF einschalten				
	TIMSK  |= (1 << TOIE0);
	clear_all();
	// Pause wg. sofort wieder getkey()
	my_delay_ms(500);
}

//************************************************************************
uint8_t scankey(volatile uint8_t *pinx, uint8_t pin) 
{
	if (bit_is_clear(*pinx, pin)) {
		my_delay_ms(DEBOUNCE_MS);
		if (bit_is_clear(*pinx, pin)) return 1;
	}
	return 0;
}

//************************************************************************
uint8_t getkey(void) 
{
	if (scankey(&TASTER_ON_OFF_PINX, TASTER_ON_OFF_PIN))	return ON_OFF;
	if (scankey(&TASTER_LEFT_PINX, TASTER_LEFT_PIN))		return LEFT;
	if (scankey(&TASTER_RIGHT_PINX, TASTER_RIGHT_PIN))		return RIGHT;
	return 0;
}

//*********************************************************************
// damit Gleichverteilung der Zufallszahlen ueber den ganzen Bereich 
// gewaehrleistet ist, muss ein wenig mehr gemacht werden, als rand() 
// aufzurufen...; Quelle: http://www.mikrocontroller.net/topic/80481
#define MAX_WERT 7
uint8_t my_rand(uint8_t basis)
{
	uint16_t x;
	while((x = rand()) >= RAND_MAX - (RAND_MAX % basis)); 
	return x % basis;
}

//*********************************************************************
// seed-Wert fuer srand() generieren; Quelle: 
// http://www.roboternetz.de/wissen/index.php/Zufallszahlen_mit_avr-gcc
unsigned int get_seed(void)
{
	uint16_t seed = 0;
	uint16_t *p = (uint16_t*) (RAMEND+1);
	extern uint16_t __heap_start;
	while (p >= &__heap_start + 1)
		seed ^= * (--p);
	return seed;
}

//************************************************************************
void print_sign(uint8_t sign, int8_t x)
{
	uint8_t i, row;
	int8_t j;
	// anzeigbares Zeichen? Wenn nein, dann Leerzeichen!
	if ((sign < 0x20) || (sign > 0x7F)) sign = 0x20;
	// jede Zeile des Fonts adressieren, lesen, ausgeben
	for(i=0; i<FONT_DY; i++) {
		row = (pgm_read_byte(&font[(sign-0x20)*FONT_DY+i]));
		// erstmal soweit nach rechts (pos. x) oder links (neg. x) schieben
		if (x<0) {
			// links schieben
			for (j=0; j>x; j--) row *= 2;
		} else {
			// rechts schieben
			for (j=0; j<x; j++) row /= 2;
		}
		// und mit led[] verodern (damit schon dastehendes erhalten bleibt)
		led[i] |= row;
	}
}

//************************************************************************
void print_text(const char * text)
{
	uint8_t text_x_len = strlen(text)*(FONT_DX+1);
	uint8_t cursor = 0;
	uint8_t idx_sign; 
	int8_t mx;
	
	while (cursor<text_x_len) {
		clear_all();
		// ...erstes sichtbares Zeichen des Textes ermitteln
		idx_sign=cursor/(FONT_DX+1);
		// ...x-Anfangsposition des ersten noch sichtbaren Zeichen in der Matrix
		mx=0-(cursor%(FONT_DX+1));
		// ...und alles sichbare ausgeben
		while (mx<MATRIX_DX) {
			// ... irgendwie muss hier noch eine Pruefung rein ...!
			print_sign(text[idx_sign], mx);
			mx += FONT_DX+1;
			idx_sign++;
		}
		// ...Spaltencursor im Gesamttext erhoehen
		cursor++;
		// ...und etwas warten
		my_delay_ms(150);
	}
	clear_all();
}

//************************************************************************
void flash_screen(void)
{
	uint8_t i;
	for (i=0; i<MATRIX_DX; i++) led[i] = ~led[i];
	my_delay_ms(70);
	for (i=0; i<MATRIX_DX; i++) led[i] = ~led[i];
}

//************************************************************************
uint8_t touch_cursor(uint8_t x)
{
	uint8_t i;
	for (i=cx; i<(cx+CURSOR_SIZE); i++) {
		if (i==x) return 1;
	}
	return 0;
}

//************************************************************************
void generate_bomb(uint8_t i)
{
	bomb[i][0]=my_rand(MATRIX_DX);	// x
	bomb[i][1]=0;			// Anfangswert y
	bomb[i][2]=my_rand(SPEED_NUMBERS);	// Zeit-Incr.
	bomb[i][3]=bomb[i][2];	// Basis Zeit-Incr.
	// Bombe das 1.Mal zeichnen
	set_led(bomb[i][0], bomb[i][1], 1);
}

//************************************************************************
void check_bomb(uint8_t i)
{
	// Bombe ein Schritt weiter?
	if (!bomb[i][2]) {
	// ja
		// alte Position loeschen
		set_led(bomb[i][0], bomb[i][1], 0);
		// naechste Zeile
		bomb[i][1]++;
		bomb[i][2] = bomb[i][3];
		// neue Position setzen, wenn noch sichtbar
		if (bomb[i][1]<MATRIX_DY) set_led(bomb[i][0], bomb[i][1], 1);
	} else {
	// nein
		bomb[i][2]--;
	}
	// Bombe auf der vorletzten oder letzten Zeile und im Cursorbereich?
	if ((bomb[i][1]==(MATRIX_DY-1))&&touch_cursor(bomb[i][0])) {
		// Positiv-Punkte hochzaehlen
		win_bomb++;
		// Bombe loeschen
		set_led(bomb[i][0], bomb[i][1], 0);
		// neue Bombe initialisieren
		generate_bomb(i);
	}
	// Bombe ueber die letzte Zeile hinaus?
	if (bomb[i][1]==MATRIX_DY) {
		// kurzes flackern der Anzeige
		flash_screen();
		// nein, Negativ-Punkte hochzaehlen
		lost_bomb++;
		// neue Bombe initialisieren
		generate_bomb(i);
	}
}

//************************************************************************
//************************************************************************
//************************************************************************
int main(void)
{
	uint8_t i;
	char s[20];
	// Hardware initialisieren...
	// ... Analogcomperator ausschalten (Strom sparen!)
	ACSR = 0x80;		
	//... Ausgaenge
	ROW_0_DDR |= (1<<ROW_0);
	ROW_1_DDR |= (1<<ROW_1);
	ROW_2_DDR |= (1<<ROW_2);
	ROW_3_DDR |= (1<<ROW_3);
	ROW_4_DDR |= (1<<ROW_4);
	ROW_5_DDR |= (1<<ROW_5);
	ROW_6_DDR |= (1<<ROW_6);
	ROW_7_DDR |= (1<<ROW_7);
	COL_0_DDR |= (1<<COL_0);
	COL_1_DDR |= (1<<COL_1);
	COL_2_DDR |= (1<<COL_2);
	COL_3_DDR |= (1<<COL_3);
	COL_4_DDR |= (1<<COL_4);
	COL_5_DDR |= (1<<COL_5);
	COL_6_DDR |= (1<<COL_6);
	COL_7_DDR |= (1<<COL_7);
	// ... Eingaenge
	TASTER_ON_OFF_PORT &= ~(1 << TASTER_ON_OFF_PIN);
	TASTER_LEFT_PORT   &= ~(1 << TASTER_LEFT_PIN);
	TASTER_RIGHT_PORT  &= ~(1 << TASTER_RIGHT_PIN);
	// ... Timer initialisieren
    TCNT0   = PRELOAD_T0;							// Preloadwert Timer0 setzen
	TCCR0  |= (1 << CS01)|(1 << CS00);				// 64
	TIMSK  |= (1 << TOIE0);							// Timer0-OVF einschalten
	// ... Interrupts einschalten
	sei();
		
	// Zufallsgenerator initialisieren
	srand(get_seed());
	// Begruessungstext ausgeben
	print_text("New Game!");
	// Cursor setzen
	draw_cursor(0);
	// Anfangsposition generieren
	for (i=0; i<MAX_BOMBS; i++) generate_bomb(i);
	// Endlosloop
	while (1){

		// Tasten abfragen/reagieren
		switch (getkey()) {
			case ON_OFF:	{	
								// wenn nach einer 1/2 Sekunde immer noch Taste
								// gedrueckt, dann Spielende, sonst nur Spielstand
								// ausgeben
								my_delay_ms(500);
								if (getkey()==ON_OFF) {
									// Sleep-Mode einschalten
									go_sleep();
									// neues Spiel
									print_text("New Game!");
									// Spielstand initialisieren
									win_bomb=0;
									lost_bomb=0;
									// Cursor setzen
									draw_cursor(cx);
									// Anfangsposition generieren
									for (i=0; i<MAX_BOMBS; i++) generate_bomb(i);
								} else {
									// Spielstand ausgeben
									sprintf(s, "%d/%d", win_bomb, (win_bomb+lost_bomb));
									print_text(s);
									// Cursor wieder setzen
									draw_cursor(cx);
									// Bomben zeichnen
									for (i=0; i<MAX_BOMBS; i++) check_bomb(i);
								}
								break;
							}
			case LEFT:		{
								// Cursor ein Schritt nach links
								draw_cursor(-1);
								break;
							}
			case RIGHT:		{
								// Cursor ein Schritt nach rechts
								draw_cursor(1);
								break;
							}
		}
		
		// naechster Spielschritt
		if (next_step) {
			next_step=0;
			for (i=0; i<MAX_BOMBS; i++) check_bomb(i);
			draw_cursor(0);
		}
	}
}
